__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import argparse

from reveal_user_classification.reveal import integration


def main():
    """
    Entry point.
    """
    # Parse arguments.
    parser = argparse.ArgumentParser()

    ####################################################################################################################
    # MongoDB connection for getting access to input.
    ####################################################################################################################
    parser.add_argument("-uri", "--mongo-uri", dest="mongo_uri",
                        help="A mongo client URI.",
                        type=str, required=True)
    parser.add_argument("-id", "--assessment-id", dest="assessment_id",
                        help="A mongo database-collection pair in the form: \"database_name.collection_name\".",
                        type=str, required=False, default="snow_tweet_storage.tweets")

    ####################################################################################################################
    # Twitter application credentials.
    ####################################################################################################################
    parser.add_argument("-tak", "--twitter-app-key", dest="twitter_app_key",
                        help="Twitter app key.",
                        type=str, required=True)
    parser.add_argument("-tas", "--twitter-app-secret", dest="twitter_app_secret",
                        help="Twitter app secret.",
                        type=str, required=True)

    ####################################################################################################################
    # RabbitMQ credentials for publishing a message.
    ####################################################################################################################
    parser.add_argument("-rmquri", "--rabbitmq-uri", dest="rabbitmq_uri",
                        help="RabbitMQ connection URI.",
                        type=str, required=True)
    parser.add_argument("-rmqq", "--rabbitmq-queue", dest="rabbitmq_queue",
                        help="RabbitMQ queue to check or create for publishing a success message.",
                        type=str, required=True)
    parser.add_argument("-rmqe", "--rabbitmq-exchange", dest="rabbitmq_exchange",
                        help="RabbitMQ exchange name.",
                        type=str, required=True)
    parser.add_argument("-rmqrk", "--rabbitmq-routing-key", dest="rabbitmq_routing_key",
                        help="RabbitMQ routing key (e.g. \"amqp://guest:guest@localhost:5672/vhost\").",
                        type=str, required=True)

    ####################################################################################################################
    # PServer credentials for storing data.
    ####################################################################################################################
    parser.add_argument("-pshn", "--pserver-host-name", dest="pserver_host_name",
                        help="The address of the machine where the PServer instance is hosted.",
                        type=str, required=False, default=None)
    parser.add_argument("-pscn", "--pserver-client-name", dest="pserver_client_name",
                        help="The PServer client name.",
                        type=str, required=False, default=None)
    parser.add_argument("-pscp", "--pserver-client-pass", dest="pserver_client_pass",
                        help="The PServer client's password.",
                        type=str, required=False, default=None)

    ####################################################################################################################
    # Parameters for specifying the input document set.
    ####################################################################################################################
    parser.add_argument("-ln", "--latest-n", dest="latest_n",
                        help="Get only the N most recent documents.",
                        type=int, required=False, default=None)
    parser.add_argument("-lts", "--lower-timestamp", dest="lower_timestamp",
                        help="Get only documents created after this UNIX timestamp.",
                        type=float, required=False, default=None)
    parser.add_argument("-uts", "--upper-timestamp", dest="upper_timestamp",
                        help="Get only documents created before this UNIX timestamp.",
                        type=float, required=False, default=None)

    ####################################################################################################################
    # Execution parameters.
    ####################################################################################################################
    parser.add_argument("-rp", "--restart-probability", dest="restart_probability",
                        help="Random walk restart probability.",
                        type=float, required=False, default=0.5)
    parser.add_argument("-nt", "--number-of-threads", dest="number_of_threads",
                        help="The number of parallel threads for feature extraction and classification.",
                        type=int, required=False, default=None)
    parser.add_argument("-nua", "--number-of-users-to-annotate", dest="number_of_users_to_annotate",
                        help="We extract keywords from twitter lists for a certain number of central users.",
                        type=int, required=False, default=90)  # Approximately 1 per minute.
    parser.add_argument("-mnl", "--max-number-of-labels", dest="max_number_of_labels",
                        help="The maximum number of topics/labels.",
                        type=int, required=False, default=500)

    ####################################################################################################################
    # Read/write destination information.
    ####################################################################################################################
    parser.add_argument("-unpcdb", "--user-network-profile-classifier-db", dest="user_network_profile_classifier_db",
                        help="We store the extracted twitter list keywords on another mongo database in the same client.",
                        type=str, required=False, default="user_network_profile_classifier_db")
    parser.add_argument("-lrf", "--local-resources-folder", dest="local_resources_folder",
                        help="We may have a certain number of twitter list data stored locally.",
                        type=str, required=False, default="/data/LocalStorage/memory/SNOW/twitter_lists/")

    # Extract arguments.
    args = parser.parse_args()

    mongo_uri = args.mongo_uri
    assessment_id = args.assessment_id

    twitter_app_key = args.twitter_app_key
    twitter_app_secret = args.twitter_app_secret

    rabbitmq_uri = args.rabbitmq_uri
    rabbitmq_queue = args.rabbitmq_queue
    rabbitmq_exchange = args.rabbitmq_exchange
    rabbitmq_routing_key = args.rabbitmq_routing_key

    pserver_host_name = args.pserver_host_name
    pserver_client_name = args.pserver_client_name
    pserver_client_pass = args.pserver_client_pass

    latest_n = args.latest_n
    lower_timestamp = args.lower_timestamp
    upper_timestamp = args.upper_timestamp

    restart_probability = args.restart_probability
    number_of_threads = args.number_of_threads
    number_of_users_to_annotate = args.number_of_users_to_annotate
    max_number_of_labels = args.max_number_of_labels

    user_network_profile_classifier_db = args.user_network_profile_classifier_db
    local_resources_folder = args.local_resources_folder

    integration.user_network_profile_classifier(mongo_uri=mongo_uri,
                                                assessment_id=assessment_id,
                                                twitter_app_key=twitter_app_key,
                                                twitter_app_secret=twitter_app_secret,
                                                rabbitmq_uri=rabbitmq_uri,
                                                rabbitmq_queue=rabbitmq_queue,
                                                rabbitmq_exchange=rabbitmq_exchange,
                                                rabbitmq_routing_key=rabbitmq_routing_key,
                                                pserver_host_name=pserver_host_name,
                                                pserver_client_name=pserver_client_name,
                                                pserver_client_pass=pserver_client_pass,
                                                latest_n=latest_n,
                                                lower_timestamp=lower_timestamp,
                                                upper_timestamp=upper_timestamp,
                                                restart_probability=restart_probability,
                                                number_of_threads=number_of_threads,
                                                number_of_users_to_annotate=number_of_users_to_annotate,
                                                max_number_of_labels=max_number_of_labels,
                                                user_network_profile_classifier_db=user_network_profile_classifier_db,
                                                local_resources_folder=local_resources_folder)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import json
import itertools
import datetime
import numpy as np
import scipy.sparse as spsp
import networkx as nx
from networkx.algorithms.link_analysis import pagerank_scipy
from collections import OrderedDict

from reveal_user_annotation.mongo.mongo_util import establish_mongo_connection
from reveal_user_annotation.mongo.preprocess_data import get_collection_documents_generator,\
    extract_graphs_from_tweets,  store_user_documents, read_user_documents_generator,\
    extract_connected_components, extract_graphs_and_lemmas_from_tweets
from reveal_user_annotation.twitter.user_annotate import decide_which_users_to_annotate,\
    fetch_twitter_lists_for_user_ids_generator, extract_user_keywords_generator, form_user_label_matrix
from reveal_user_annotation.pserver.request import delete_features, add_features, insert_user_data
from reveal_user_annotation.rabbitmq.rabbitmq_util import simple_notification, simpler_notification
from reveal_graph_embedding.embedding.arcte.arcte import arcte
from reveal_graph_embedding.embedding.common import normalize_columns
from reveal_graph_embedding.embedding.community_weighting import chi2_psnr_community_weighting
from reveal_graph_embedding.learning.classification import model_fit, classify_users


def make_time_window_filter(lower_timestamp, upper_timestamp):
    if lower_timestamp is not None:
        # lower_datetime = datetime.datetime.strftime(datetime.datetime.utcfromtimestamp(lower_timestamp),
        #                                             "%b %d %Y %H:%M:%S")
        lower_datetime = datetime.datetime.utcfromtimestamp(lower_timestamp)
        if upper_timestamp is not None:
            # upper_datetime = datetime.datetime.strftime(datetime.datetime.utcfromtimestamp(upper_timestamp),
            #                                             "%b %d %Y %H:%M:%S")
            upper_datetime = datetime.datetime.utcfromtimestamp(upper_timestamp)
            # Both timestamps are defined.
            spec = dict()
            spec["created_at"] = {"$gte": lower_datetime, "$lt": upper_datetime}
        else:
            spec = dict()
            spec["created_at"] = {"$gte": lower_datetime}
    else:
        if upper_timestamp is not None:
            # upper_datetime = datetime.datetime.strftime(datetime.datetime.utcfromtimestamp(upper_timestamp),
            #                                             "%b %d %Y %H:%M:%S")
            upper_datetime = datetime.datetime.utcfromtimestamp(upper_timestamp)
            spec = dict()
            spec["created_at"] = {"$lt": upper_datetime}
        else:
            spec = None
    return spec


def safe_establish_mongo_connection(mongo_uri, assessment_id):
    """
           - tweet_input_database_name: The mongo database name where the input tweets are stored.
           - tweet_input_collection_name: The mongo collection name where the input tweets are stored.
    """

    client = establish_mongo_connection(mongo_uri)

    tweet_input_database_name, tweet_input_collection_name = translate_assessment_id(assessment_id)

    return client, tweet_input_database_name, tweet_input_collection_name


def translate_assessment_id(assessment_id):
    """
    The assessment id is translated to MongoDB database and collection names.

    Input:   - assessment_id: The connection details for making a connection with a MongoDB instance.

    Outputs: - database_name: The name of the Mongo database in string format.
             - collection_name: The name of the collection of tweets to read in string format.
    """
    assessment_id = assessment_id.split(".")

    database_name = assessment_id[0]
    collection_name = assessment_id[1]

    return database_name, collection_name


def get_graphs_and_lemma_matrix(client,
                                tweet_input_database_name,
                                tweet_input_collection_name,
                                spec,
                                latest_n):
    """
    Processes a set of tweets and extracts interaction graphs and a user-lemma vector representation matrix.

    Inputs:  - client: A MongoDB client.
             - tweet_input_database_name: The name of a Mongo database.
             - tweet_input_collection_name: The name of the collection of tweets.
             - spec: A python dictionary that defines higher query arguments.
             - latest_n: The number of latest results we require from the mongo document collection.

    Outputs: - mention_graph: The mention graph as a SciPy sparse matrix.
             - retweet_graph: The retweet graph as a SciPy sparse matrix.
             - user_lemma_matrix: The user lemma vector representation matrix as a SciPy sparse matrix.
             - tweet_id_set: A python set containing the Twitter ids for all the dataset tweets.
             - user_id_set: A python set containing the Twitter ids for all the dataset users.
             - lemma_to_attribute: A map from lemmas to numbers in python dictionary format.
    """
    # Form the mention and retweet graphs, as well as the attribute matrix.
    tweet_gen = get_collection_documents_generator(client=client,
                                                   database_name=tweet_input_database_name,
                                                   collection_name=tweet_input_collection_name,
                                                   spec=spec,
                                                   latest_n=latest_n,
                                                   sort_key="created_at")

    mention_graph,\
    retweet_graph,\
    tweet_id_set,\
    user_id_set,\
    node_to_id,\
    id_to_name = extract_graphs_from_tweets(tweet_gen)

    return mention_graph, retweet_graph, user_id_set, node_to_id


def integrate_graphs(mention_graph, retweet_graph, node_to_id, restart_probability, number_of_threads):
    """
    A bit of post-processing of the graphs to end up with a single aggregate graph.

    Inputs:  - mention_graph: The mention graph as a SciPy sparse matrix.
             - retweet_graph: The retweet graph as a SciPy sparse matrix.
             - user_lemma_matrix: The user lemma vector representation matrix as a SciPy sparse matrix.
             - number_of_threads:

    Outputs: - adjacency_matrix: An aggregate, post-processed view of the graphs.
             - node_to_id: A node to Twitter id map as a python dictionary.
             - features: The graph structure proximity features as calculated by ARCTE in scipy sparse matrix format.
             - node_importances: A vector containing node importance values.
    """
    # Form the adjacency matrix.
    adjacency_matrix = 0.25*mention_graph +\
                       0.25*mention_graph.transpose() +\
                       0.25*retweet_graph +\
                       0.25*retweet_graph.transpose()

    # Here is where I need to extract connected components or something similar.
    adjacency_matrix, node_to_id, old_node_list = extract_connected_components(adjacency_matrix, "weak", node_to_id)

    # Extract features
    features = arcte(adjacency_matrix=adjacency_matrix,
                     rho=restart_probability,
                     epsilon=0.00001,
                     number_of_threads=number_of_threads)

    node_importances = calculate_node_importances(adjacency_matrix)

    return adjacency_matrix, node_to_id, features, node_importances


def calculate_node_importances(adjacency_matrix):
    """
    Calculates a vector that contains node importance values. This nodes are to be automatically annotated later.

    Input:  - adjacency_matrix

    Output: - node_importances: A vector containing node importance values.
    """
    graph = nx.from_scipy_sparse_matrix(adjacency_matrix, create_using=nx.Graph())

    node_importances = pagerank_scipy(graph, alpha=0.9, max_iter=500, tol=1.0e-8)

    node_importances = OrderedDict(sorted(node_importances.items(), key=lambda t: t[0]))

    node_importances = np.array(list(node_importances.values()), dtype=np.float64)

    return node_importances


def fetch_twitter_lists(client,
                        twitter_app_key,
                        twitter_app_secret,
                        user_network_profile_classifier_db,
                        local_resources_folder,
                        centrality,
                        number_of_users_to_annotate,
                        node_to_id):
    """
    Decides which users to annotate and fetcher Twitter lists as needed.

    Inputs:  - client: A MongoDB client.
             - centrality: A vector containing centrality measure values.
             - node_to_id: A node to Twitter id map as a python dictionary.

    Outputs: - twitter_lists_gen: A python generator that generates Twitter list generators.
             - user_ids_to_annotate: A list of Twitter user ids.
    """
    user_twitter_ids_mongo, user_twitter_ids_local = find_already_annotated(client=client,
                                                                            mongo_database_name=user_network_profile_classifier_db,
                                                                            local_resources_folder=local_resources_folder)

    already_annotated_user_ids = (set(user_twitter_ids_mongo + user_twitter_ids_local))

    # Calculate the most central users.
    if number_of_users_to_annotate > 0:
        user_ids_to_annotate = decide_which_users_to_annotate(centrality_vector=centrality,
                                                              number_to_annotate=number_of_users_to_annotate,
                                                              already_annotated=already_annotated_user_ids,
                                                              node_to_id=node_to_id)
    else:
        user_ids_to_annotate = list()

    # Fetch Twitter lists.
    twitter_lists_gen = fetch_twitter_lists_for_user_ids_generator(twitter_app_key,
                                                                   twitter_app_secret,
                                                                   user_ids_to_annotate)

    # # Store Twitter lists in MongoDB.
    # store_user_documents(twitter_lists_gen,
    #                      client=client,
    #                      database_name="twitter_list_database")
    #
    #
    # twitter_lists_gen = read_user_documents_generator(user_ids_to_annotate,
    #                                                   client=client,
    #                                                   database_name="twitter_list_database")

    return twitter_lists_gen, user_ids_to_annotate, user_twitter_ids_mongo, user_twitter_ids_local


def find_already_annotated(client, mongo_database_name, local_resources_folder):
    """
    Finds the twitter ids of the users that have already been annotated.

    Inputs:  - client: A MongoDB client.
             - mongo_database_name: The name of a Mongo database as a string.

    Outputs: - user_twitter_ids_mongo: A python list of user twitter ids already annotated.
             - user_twitter_ids_local: A python list of user twitter ids already annotated.

    """
    # Check the mongo database for user annotation.
    db = client[mongo_database_name]
    collection = db["twitter_list_keywords_collection"]

    cursor = collection.find()

    user_twitter_ids_mongo = list()
    append_user_twitter_user_id = user_twitter_ids_mongo.append
    for document in cursor:
        append_user_twitter_user_id(document["_id"])

    # Check locally for user annotation.
    if local_resources_folder is not None:
        file_list = os.listdir(local_resources_folder)
        user_twitter_ids_local = [int(user_twitter_id[:-5]) for user_twitter_id in file_list]
    else:
        user_twitter_ids_local = list()

    return user_twitter_ids_mongo, user_twitter_ids_local


def annotate_users(client, twitter_lists_gen,
                   user_ids_to_annotate,
                   user_twitter_ids_mongo,
                   user_twitter_ids_local,
                   local_resources_folder,
                   user_network_profile_classifier_db,
                   node_to_id,
                   max_number_of_labels):
    """
    Forms a user-to-label matrix by annotating certain users.

    Inputs:  - client: A MongoDB client.
             - twitter_lists_gen: A python generator that generates Twitter list generators.
             - user_ids_to_annotate: A list of Twitter user ids.
             - node_to_id: A node to Twitter id map as a python dictionary.

    Outputs: - user_label_matrix: A user-to-label matrix in scipy sparse matrix format.
             - annotated_user_ids: A list of Twitter user ids.
    """
    # Process lists and store keywords in mongo.
    # TODO: Do asynchronous I/O and preprocessing.
    user_twitter_list_keywords_gen = extract_user_keywords_generator(twitter_lists_gen,
                                                                     lemmatizing="wordnet")

    store_user_documents(user_twitter_list_keywords_gen,
                         client=client,
                         mongo_database_name=user_network_profile_classifier_db,
                         mongo_collection_name="twitter_list_keywords_collection")

    # Read local resources as well.
    # Calculate which user annotations to fetch.
    user_twitter_ids_local = np.intersect1d(np.array(list(node_to_id.values()), dtype=int), np.array(user_twitter_ids_local, dtype=int))
    local_user_twitter_list_keywords_gen = read_local_user_annotations(json_folder=local_resources_folder,
                                                                       user_twitter_ids=user_twitter_ids_local)

    # Calculate which user annotations to fetch.
    user_ids_to_fetch = np.intersect1d(np.array(list(node_to_id.values()), dtype=int), np.array(user_twitter_ids_mongo, dtype=int))

    mongo_user_twitter_list_keywords_gen = read_user_documents_generator(user_ids_to_fetch,
                                                                         client=client,
                                                                         mongo_database_name=user_network_profile_classifier_db,
                                                                         mongo_collection_name="twitter_list_keywords_collection")

    user_twitter_list_keywords_gen = itertools.chain(local_user_twitter_list_keywords_gen,
                                                     mongo_user_twitter_list_keywords_gen)

    # Annotate users.
    id_to_node = dict(zip(node_to_id.values(), node_to_id.keys()))
    user_label_matrix, annotated_user_ids, label_to_lemma, lemma_to_keyword = form_user_label_matrix(user_twitter_list_keywords_gen,
                                                                                                     id_to_node,
                                                                                                     max_number_of_labels)

    return user_label_matrix, annotated_user_ids, label_to_lemma, lemma_to_keyword


def read_local_user_annotations(json_folder,
                                user_twitter_ids):
    if json_folder is not None:
        for user_twitter_id in user_twitter_ids:
            path = json_folder + "/" + str(user_twitter_id) + ".json"
            with open(path, "r", encoding="utf-8") as f:
                twitter_lists = json.load(f)

                yield user_twitter_id, twitter_lists
    else:
        raise StopIteration


def user_classification(features, user_label_matrix, annotated_user_ids, node_to_id, number_of_threads):
    """
    Perform user classification.

    Inputs:  - features: The graph structure proximity features as calculated by ARCTE in scipy sparse matrix format.
             - user_label_matrix: A user-to-label matrix in scipy sparse matrix format.
             - annotated_user_ids: A list of Twitter user ids.
             - node_to_id: A node to Twitter id map as a python dictionary.
             - number_of_threads:

    Output:  - prediction: The output of the classification in scipy sparse matrix format.
    """
    non_annotated_user_ids = np.setdiff1d(np.arange(len(node_to_id), dtype=int), annotated_user_ids)

    features = normalize_columns(features)

    X_train = features[annotated_user_ids, :]
    X_test = features[non_annotated_user_ids, :]
    y_train = user_label_matrix[annotated_user_ids, :]

    X_train, X_test = chi2_psnr_community_weighting(X_train, X_test, y_train)
    print("Performed community weighting.")

    model = model_fit(X_train,
                      y_train,
                      svm_hardness=1.0,
                      fit_intercept=True,
                      number_of_threads=number_of_threads,
                      # classifier_type="LogisticRegression")
                      classifier_type="RandomForest")
    prediction = spsp.csr_matrix(user_label_matrix.shape, dtype=np.float64)
    y_pred = classify_users(X_test,
                            model,
                            # classifier_type="LogisticRegression")
                            classifier_type="RandomForest")
    y_pred = spsp.csr_matrix(y_pred)
    prediction[non_annotated_user_ids, :] = y_pred
    prediction[annotated_user_ids, :] = user_label_matrix[annotated_user_ids, :]
    prediction.eliminate_zeros()

    return prediction


def get_user_topic_generator(prediction, node_to_id, label_to_lemma, lemma_to_keyword):
    """
    Generates twitter user ids along with their associated topic strings.

    Inputs: - prediction: A scipy sparse matrix that has non-zero values in cases a node is associated with a label.
            - node_to_id: A node to Twitter id map as a python dictionary.
            - label_to_lemma: A map from numbers to string lemmas in python dictionary format.
            - lemma_to_keyword: A map from lemmas to original keyword in python dictionary format.

    Yields: - twitter_user_id: A twitter user id in integer format.
            - topics: A generator of topic strings.
    """
    number_of_users = prediction.shape[0]

    prediction = spsp.csr_matrix(prediction)
    for node in range(number_of_users):
        twitter_user_id = node_to_id[node]

        prediction_row = prediction.getrow(node)
        labels = list(prediction_row.indices)
        label_scores = list(prediction_row.data)
        labels_and_label_scores = zip(labels, label_scores)

        if len(labels) != 0:
            topics = dict()
            topics["topic_to_score"] = {lemma_to_keyword[label_to_lemma[label]]: score for label, score in labels_and_label_scores}

            yield twitter_user_id, topics


def write_results_to_mongo(client,
                           user_network_profile_classifier_db,
                           user_topic_gen):
    """
    What it says on the tin.

    Inputs: - client: A MongoDB client.
            - user_network_profile_classifier_db:
            - user_topic_gen: A python generator that generates users and a generator of associated topic keywords.
    """
    store_user_documents(user_topic_gen,
                         client=client,
                         mongo_database_name=user_network_profile_classifier_db,
                         mongo_collection_name="user_topics_collection")


def write_topics_to_pserver(host_name,
                            client_name,
                            client_pass,
                            user_topic_gen,
                            topic_list):
    """
    What is says on the tin.

    Inputs: - host_name: A string containing the address of the machine where the PServer instance is hosted.
            - client_name: The PServer client name.
            - client_pass: The PServer client's password.
            - user_topic_gen: A python generator that generates users and a generator of associated topic keywords.
    """
    # Make sure the database exists.
    delete_features(host_name=host_name,
                    client_name=client_name,
                    client_pass=client_pass,
                    feature_names=None)

    topic_list = ["type." + topic for topic in topic_list]
    add_features(host_name=host_name,
                 client_name=client_name,
                 client_pass=client_pass,
                 feature_names=topic_list)

    # Store the score-label tuples.
    for user_twitter_id, topic_to_score in user_topic_gen:
        insert_user_data(host_name=host_name,
                         client_name=client_name,
                         client_pass=client_pass,
                         user_twitter_id=user_twitter_id,
                         topic_to_score=topic_to_score)


def publish_results_via_rabbitmq(rabbitmq_connection,
                                 rabbitmq_queue,
                                 rabbitmq_exchange,
                                 rabbitmq_routing_key,
                                 user_topic_gen):
    """
    What is says on the tin.

    Inputs:
            - user_topic_gen: A python generator that generates users and a generator of associated topic keywords.
    """
    # for user_twitter_id, topic_to_score in user_topic_gen:
    #     print(user_twitter_id, topic_to_score)

    channel = rabbitmq_connection.channel()
    channel.queue_declare(rabbitmq_queue, durable=True)
    channel.exchange_declare(rabbitmq_exchange, type='direct', durable=True)
    channel.queue_bind(rabbitmq_queue, rabbitmq_exchange, routing_key=rabbitmq_routing_key)

    for user_twitter_id, topic_to_score in user_topic_gen:
        results_dictionary = dict()
        results_dictionary["contributor_id"] = str(user_twitter_id)
        results_dictionary["type"] = [[str(user_type), repr(score)] for user_type, score in topic_to_score.items()]

        results_string = json.dumps(results_dictionary)

        # Publish the message.
        simpler_notification(channel,
                             rabbitmq_queue,
                             rabbitmq_exchange,
                             rabbitmq_routing_key,
                             results_string)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from reveal_user_classification.preprocess.insight.insight_curation_util import make_folder_paths,\
    get_number_of_nodes, preprocess_graph_data, preprocess_feature_data, make_implicit_graphs, make_labelling


def preprocess_insight_curation_dataset(insight_curation_source_folder,
                                        dataset,
                                        graph_raw_data_files,
                                        feature_raw_data_files):
    # Calculate folder paths.
    raw_data_folder,\
    preprocessed_data_folder,\
    implicit_graph_folder,\
    simple_undirected_implicit_graph_folder = make_folder_paths(insight_curation_source_folder, dataset)

    # Get number of nodes.
    number_of_nodes = get_number_of_nodes(raw_data_folder, dataset)

    # Preprocess all graph data.
    for graph_raw_data_file in graph_raw_data_files:
        preprocess_graph_data(dataset,
                              raw_data_folder,
                              preprocessed_data_folder,
                              graph_raw_data_file,
                              number_of_nodes)

    # Preprocess all feature data.
    for feature_raw_data_file in feature_raw_data_files:
        preprocess_feature_data(dataset,
                                raw_data_folder,
                                preprocessed_data_folder,
                                feature_raw_data_file,
                                number_of_nodes)

    # Make all implicit graphs.
    make_implicit_graphs(preprocessed_data_folder, simple_undirected_implicit_graph_folder)

    # Make labelling.
    make_labelling(dataset, raw_data_folder, preprocessed_data_folder)


########################################################################################################################
# Configure preprocessing.
########################################################################################################################
SOURCE_FOLDER = "/home/georgerizos/Documents/LocalStorage/raw_data/Insight/Curation"

CURATION_DATASETS = ["alaska", "georgia", "idaho", "massachusetts", "northdakota", "ohio", "oklahoma", "tennessee",
                     "vermont", "virginia"]

GRAPH_RAW_DATA_FILES = ["followedby", "mentionedby", "retweetedby"]

FEATURE_RAW_DATA_FILES = ["colisted", "listdesc", "listmerged", "listnames", "tweets50", "tweets100", "tweets200"]

########################################################################################################################
# Preprocess all datasets.
########################################################################################################################
for DATASET in CURATION_DATASETS:
    preprocess_insight_curation_dataset(SOURCE_FOLDER,
                                        DATASET,
                                        GRAPH_RAW_DATA_FILES,
                                        FEATURE_RAW_DATA_FILES)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import argparse
import os
import numpy as np
from sklearn.multiclass import OneVsRestClassifier
from sklearn import svm

from reveal_user_classification.common import get_threads_number
from reveal_graph_embedding.datautil.score_rw_util import write_results
from reveal_graph_embedding.datautil.snow_datautil.snow_read_data import read_adjacency_matrix,\
    read_node_label_matrix
from reveal_graph_embedding.embedding.arcte.arcte import arcte
from reveal_graph_embedding.embedding.common import normalize_columns
from reveal_graph_embedding.embedding.community_weighting import chi2_contingency_matrix,\
    peak_snr_weight_aggregation, community_weighting
from reveal_graph_embedding.learning.holdout import generate_folds
from reveal_graph_embedding.learning.evaluation import form_node_label_prediction_matrix
from reveal_graph_embedding.learning import evaluation


def main():
    """
    Entry point.
    """
    # Parse arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument("-input", "--snow-tweets-folder", dest="snow_tweets_folder",
                        help="Folder that contains the SNOW tweets.",
                        type=str, required=True,)
    parser.add_argument("-output", "--prototype-output-folder", dest="prototype_output_folder",
                        help="Folder to which the results from the execution will be written.",
                        type=str, required=True)

    parser.add_argument("-rp", "--restart-probability", dest="restart_probability",
                        help="Random walk restart probability.",
                        type=float, required=False, default=0.2)
    parser.add_argument("-nt", "--number-of-threads", dest="number_of_threads",
                        help="The number of parallel threads for feature extraction and classification.",
                        type=int, required=False, default=None)

    args = parser.parse_args()

    snow_tweets_folder = args.snow_tweets_folder
    prototype_output_folder = args.prototype_output_folder

    restart_probability = args.restart_probability
    number_of_threads = args.number_of_threads

    run_prototype(snow_tweets_folder=snow_tweets_folder,
                  prototype_output_folder=prototype_output_folder,
                  restart_probability=restart_probability,
                  number_of_threads=number_of_threads)


def run_prototype(snow_tweets_folder,
                  prototype_output_folder,
                  restart_probability,
                  number_of_threads):
    """
    This is a sample execution of the User Network Profile Classifier Prototype.

    Specifically:
           - Reads a set of tweets from a local folder.
           - Forms graphs and text-based vector representation for the users involved.
           - Fetches Twitter lists for influential users.
           - Extracts keywords from Twitter lists and thus annotates these users as experts in these topics.
           - Extracts graph-based features using the ARCTE algorithm.
           - Performs user classification for the rest of the users.
    """
    if number_of_threads is None:
        number_of_threads = get_threads_number()

    ####################################################################################################################
    # Read data.
    ####################################################################################################################
    # Read graphs.
    edge_list_path = os.path.normpath(snow_tweets_folder + "/graph.tsv")
    adjacency_matrix = read_adjacency_matrix(file_path=edge_list_path,
                                             separator='\t')
    number_of_nodes = adjacency_matrix.shape[0]

    # Read labels.
    node_label_list_path = os.path.normpath(snow_tweets_folder + "/user_label_matrix.tsv")
    user_label_matrix, number_of_categories, labelled_node_indices = read_node_label_matrix(node_label_list_path,
                                                                                            '\t')

    ####################################################################################################################
    # Extract features.
    ####################################################################################################################
    features = arcte(adjacency_matrix,
                     restart_probability,
                     0.00001,
                     number_of_threads=number_of_threads)

    features = normalize_columns(features)

    percentages = np.arange(1, 11, dtype=np.int)
    trial_num = 10

    ####################################################################################################################
    # Perform user classification.
    ####################################################################################################################
    mean_macro_precision = np.zeros(percentages.size, dtype=np.float)
    std_macro_precision = np.zeros(percentages.size, dtype=np.float)
    mean_micro_precision = np.zeros(percentages.size, dtype=np.float)
    std_micro_precision = np.zeros(percentages.size, dtype=np.float)
    mean_macro_recall = np.zeros(percentages.size, dtype=np.float)
    std_macro_recall = np.zeros(percentages.size, dtype=np.float)
    mean_micro_recall = np.zeros(percentages.size, dtype=np.float)
    std_micro_recall = np.zeros(percentages.size, dtype=np.float)
    mean_macro_F1 = np.zeros(percentages.size, dtype=np.float)
    std_macro_F1 = np.zeros(percentages.size, dtype=np.float)
    mean_micro_F1 = np.zeros(percentages.size, dtype=np.float)
    std_micro_F1 = np.zeros(percentages.size, dtype=np.float)
    F1 = np.zeros((percentages.size, number_of_categories), dtype=np.float)
    for p in np.arange(percentages.size):
        percentage = percentages[p]
        # Initialize the metric storage arrays to zero
        macro_precision = np.zeros(trial_num, dtype=np.float)
        micro_precision = np.zeros(trial_num, dtype=np.float)
        macro_recall = np.zeros(trial_num, dtype=np.float)
        micro_recall = np.zeros(trial_num, dtype=np.float)
        macro_F1 = np.zeros(trial_num, dtype=np.float)
        micro_F1 = np.zeros(trial_num, dtype=np.float)
        trial_F1 = np.zeros((trial_num, number_of_categories), dtype=np.float)

        folds = generate_folds(user_label_matrix,
                               labelled_node_indices,
                               number_of_categories,
                               percentage,
                               trial_num)
        for trial in np.arange(trial_num):
            train, test = next(folds)
            ########################################################################################################
            # Separate train and test sets
            ########################################################################################################
            X_train, X_test, y_train, y_test = features[train, :],\
                                                features[test, :],\
                                                user_label_matrix[train, :],\
                                                user_label_matrix[test, :]

            contingency_matrix = chi2_contingency_matrix(X_train, y_train)
            community_weights = peak_snr_weight_aggregation(contingency_matrix)
            X_train, X_test = community_weighting(X_train, X_test, community_weights)

            ####################################################################################################
            # Train model
            ####################################################################################################
            # Train classifier
            model = OneVsRestClassifier(svm.LinearSVC(C=1,
                                                      random_state=None,
                                                      dual=False,
                                                      fit_intercept=True),
                                        n_jobs=number_of_threads)

            model.fit(X_train, y_train)
            ####################################################################################################
            # Make predictions
            ####################################################################################################
            y_pred = model.decision_function(X_test)

            y_pred = form_node_label_prediction_matrix(y_pred, y_test)

            ########################################################################################################
            # Calculate measures
            ########################################################################################################
            measures = evaluation.calculate_measures(y_pred, y_test)

            macro_recall[trial] = measures[0]
            micro_recall[trial] = measures[1]

            macro_precision[trial] = measures[2]
            micro_precision[trial] = measures[3]

            macro_F1[trial] = measures[4]
            micro_F1[trial] = measures[5]

            trial_F1[trial, :] = measures[6]

        mean_macro_precision[p] = np.mean(macro_precision)
        std_macro_precision[p] = np.std(macro_precision)
        mean_micro_precision[p] = np.mean(micro_precision)
        std_micro_precision[p] = np.std(micro_precision)
        mean_macro_recall[p] = np.mean(macro_recall)
        std_macro_recall[p] = np.std(macro_recall)
        mean_micro_recall[p] = np.mean(micro_recall)
        std_micro_recall[p] = np.std(micro_recall)
        mean_macro_F1[p] = np.mean(macro_F1)
        std_macro_F1[p] = np.std(macro_F1)
        mean_micro_F1[p] = np.mean(micro_F1)
        std_micro_F1[p] = np.std(micro_F1)
        F1[p, :] = np.mean(trial_F1, axis=0)

    measure_list = [(mean_macro_precision, std_macro_precision),
                    (mean_micro_precision, std_micro_precision),
                    (mean_macro_recall, std_macro_recall),
                    (mean_micro_recall, std_micro_recall),
                    (mean_macro_F1, std_macro_F1),
                    (mean_micro_F1, std_micro_F1),
                    F1]

    write_results(measure_list,
                  os.path.normpath(prototype_output_folder + "/F1_average_scores.txt"))



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import gc
import numpy as np
from scipy import sparse as spsp

from reveal_graph_embedding.common import get_file_row_generator
from reveal_graph_embedding.datautil.insight_datautil.insight_read_data import scipy_sparse_to_csv,\
    read_adjacency_matrix
from reveal_graph_embedding.datautil.make_directory_tree import make_sure_path_exists


def make_folder_paths(insight_curation_source_folder, dataset):
    # Make dataset raw data folder path.
    raw_data_folder = insight_curation_source_folder + "/" + dataset
    preprocessed_data_folder = raw_data_folder + "/preprocessed"
    implicit_graph_folder = preprocessed_data_folder + "/implicit_graph"
    simple_undirected_implicit_graph_folder = implicit_graph_folder + "/simple_undirected_implicit_graph"

    # Make sure dataset preprocessed data folder path exists.
    make_sure_path_exists(preprocessed_data_folder)
    make_sure_path_exists(implicit_graph_folder)
    make_sure_path_exists(simple_undirected_implicit_graph_folder)

    return raw_data_folder, preprocessed_data_folder, implicit_graph_folder, simple_undirected_implicit_graph_folder


def get_number_of_nodes(raw_data_folder, dataset):
    node_file_path = raw_data_folder + "/" + dataset + ".ids"

    file_row_gen = get_file_row_generator(node_file_path, " ")

    number_of_nodes = 0

    for file_row in file_row_gen:
        if file_row[0] == "":
            break
        else:
            number_of_nodes += 1

    return number_of_nodes


def read_graph_raw_data_file(filepath, number_of_nodes):

    file_row_gen = get_file_row_generator(filepath, " ")

    file_row = next(file_row_gen)
    while file_row[0][0] == "%":
        file_row = next(file_row_gen)

    number_of_edges = int(file_row[2])

    row = np.empty(number_of_edges, dtype=np.int32)
    col = np.empty(number_of_edges, dtype=np.int32)
    data = np.empty(number_of_edges, dtype=np.float64)

    edge_counter = 0
    for file_row in file_row_gen:
        if file_row[0] == "":
            break
        source_node = int(file_row[0])
        target_node = int(file_row[1])
        edge_weight = float(file_row[2])

        row[edge_counter] = source_node
        col[edge_counter] = target_node
        data[edge_counter] = edge_weight

        edge_counter += 1

    row = row - 1
    col = col - 1

    matrix = spsp.coo_matrix((data, (row, col)), shape=(number_of_nodes, number_of_nodes))
    matrix = spsp.coo_matrix(spsp.csr_matrix(matrix))

    return matrix


def read_feature_raw_data_file(filepath, number_of_nodes):

    file_row_gen = get_file_row_generator(filepath, " ")

    file_row = next(file_row_gen)
    while file_row[0][0] == "%":
        file_row = next(file_row_gen)

    number_of_features = int(file_row[0])
    number_of_edges = int(file_row[2])

    row = np.empty(number_of_edges, dtype=np.int32)
    col = np.empty(number_of_edges, dtype=np.int32)
    data = np.empty(number_of_edges, dtype=np.float64)

    edge_counter = 0
    for file_row in file_row_gen:
        if file_row[0] == "":
            break
        source_node = int(file_row[0])
        target_node = int(file_row[1])
        edge_weight = float(file_row[2])

        row[edge_counter] = source_node
        col[edge_counter] = target_node
        data[edge_counter] = edge_weight

        edge_counter += 1

    row = row - 1
    col = col - 1

    matrix = spsp.coo_matrix((data, (row, col)), shape=(number_of_features, number_of_nodes))
    matrix = spsp.coo_matrix(spsp.csr_matrix(matrix))

    return matrix


def preprocess_graph_data(dataset, raw_data_folder, preprocessed_data_folder, graph_raw_data_file, number_of_nodes):
    source_file_path = raw_data_folder + "/" + dataset + "-" + graph_raw_data_file + ".mtx"
    target_file_path = preprocessed_data_folder + "/" + graph_raw_data_file + ".tsv"

    matrix = read_graph_raw_data_file(source_file_path, number_of_nodes)

    scipy_sparse_to_csv(target_file_path, matrix, separator="\t", directed=True, numbering="matlab")


def preprocess_feature_data(dataset, raw_data_folder, preprocessed_data_folder, feature_raw_data_file, number_of_nodes):
    source_file_path = raw_data_folder + "/" + dataset + "-" + feature_raw_data_file + ".mtx"
    target_file_path = preprocessed_data_folder + "/" + feature_raw_data_file + ".tsv"

    matrix = read_feature_raw_data_file(source_file_path, number_of_nodes)

    scipy_sparse_to_csv(target_file_path, matrix, separator="\t", directed=True, numbering="matlab")


def make_implicit_graphs(preprocessed_data_folder, simple_undirected_implicit_graph_folder):
    ####################################################################################################################
    # Read graphs.
    ####################################################################################################################
    # Read follow graph.
    source_path = preprocessed_data_folder + "/" + "followedby" + ".tsv"
    follow_graph = read_adjacency_matrix(source_path, "\t", "matlab")
    follow_graph = follow_graph.transpose()

    # Read mention graph.
    source_path = preprocessed_data_folder + "/" + "mentionedby" + ".tsv"
    mention_graph = read_adjacency_matrix(source_path, "\t", "matlab")
    mention_graph = mention_graph.transpose()

    # Read retweet graph.
    source_path = preprocessed_data_folder + "/" + "retweetedby" + ".tsv"
    retweet_graph = read_adjacency_matrix(source_path, "\t", "matlab")
    retweet_graph = retweet_graph.transpose()

    ####################################################################################################################
    # Simple undirected implicit graphs.
    ####################################################################################################################
    target_path = simple_undirected_implicit_graph_folder + "/" + "follow_graph" + ".tsv"
    simple_undirected_follow_graph = (follow_graph + follow_graph.transpose())/2
    scipy_sparse_to_csv(target_path, simple_undirected_follow_graph, separator="\t", directed=False, numbering="matlab")

    target_path = simple_undirected_implicit_graph_folder + "/" + "mention_graph" + ".tsv"
    simple_undirected_mention_graph = (mention_graph + mention_graph.transpose())/2
    scipy_sparse_to_csv(target_path, simple_undirected_mention_graph, separator="\t", directed=False, numbering="matlab")

    target_path = simple_undirected_implicit_graph_folder + "/" + "retweet_graph" + ".tsv"
    simple_undirected_retweet_graph = (retweet_graph + retweet_graph.transpose())/2
    scipy_sparse_to_csv(target_path, simple_undirected_retweet_graph, separator="\t", directed=False, numbering="matlab")

    gc.collect()

    ####################################################################################################################
    # Multiview graphs.
    ####################################################################################################################
    target_path = simple_undirected_implicit_graph_folder + "/" + "fol_men_graph" + ".tsv"
    simple_undirected_fol_men_graph = (follow_graph + follow_graph.transpose() +
                                       mention_graph + mention_graph.transpose())/4
    scipy_sparse_to_csv(target_path, simple_undirected_fol_men_graph, separator="\t", directed=False, numbering="matlab")

    target_path = simple_undirected_implicit_graph_folder + "/" + "men_ret_graph" + ".tsv"
    simple_undirected_men_ret_graph = (mention_graph + mention_graph.transpose() +
                                       retweet_graph + retweet_graph.transpose())/4
    scipy_sparse_to_csv(target_path, simple_undirected_men_ret_graph, separator="\t", directed=False, numbering="matlab")

    target_path = simple_undirected_implicit_graph_folder + "/" + "fol_ret_graph" + ".tsv"
    simple_undirected_fol_ret_graph = (follow_graph + follow_graph.transpose() +
                                       retweet_graph + retweet_graph.transpose())/4
    scipy_sparse_to_csv(target_path, simple_undirected_fol_ret_graph, separator="\t", directed=False, numbering="matlab")

    target_path = simple_undirected_implicit_graph_folder + "/" + "fol_men_ret_graph" + ".tsv"
    simple_undirected_fol_men_ret_graph = (follow_graph + follow_graph.transpose() +
                                           mention_graph + mention_graph.transpose() +
                                           retweet_graph + retweet_graph.transpose())/6
    scipy_sparse_to_csv(target_path, simple_undirected_fol_men_ret_graph, separator="\t", directed=False, numbering="matlab")


def make_labelling(dataset, raw_data_folder, preprocessed_data_folder):
    node_file_path = raw_data_folder + "/" + dataset + ".ids"

    file_row_gen = get_file_row_generator(node_file_path, " ")

    user_twitter_id_list = list()

    for file_row in file_row_gen:
        if file_row[0] == "":
            break
        else:
            user_twitter_id_list.append(int(file_row[0]))

    id_to_node = dict(zip(user_twitter_id_list, range(len(user_twitter_id_list))))
    user_twitter_id_list = set(user_twitter_id_list)

    core_file_path = raw_data_folder + "/" + dataset + ".core"

    file_row_gen = get_file_row_generator(core_file_path, " ")

    core_user_twitter_id_list = list()

    for file_row in file_row_gen:
        if file_row[0] == "":
            break
        else:
            core_user_twitter_id_list.append(int(file_row[0]))

    core_user_twitter_id_list = user_twitter_id_list.intersection(core_user_twitter_id_list)

    non_core_user_twitter_id_set = user_twitter_id_list.difference(core_user_twitter_id_list)

    row = [id_to_node[id] for id in core_user_twitter_id_list] + [id_to_node[id] for id in non_core_user_twitter_id_set]
    row = np.array(row, dtype=np.int32)
    col = [1 for id in core_user_twitter_id_list] + [0 for id in non_core_user_twitter_id_set]
    col = np.array(col, dtype=np.int32)
    data = np.ones(len(user_twitter_id_list), dtype=np.int8)

    node_label_matrix = spsp.coo_matrix((data, (row, col)), shape=(len(user_twitter_id_list), 2))

    target_path = preprocessed_data_folder + "/" + "node_label_matrix" + ".tsv"
    scipy_sparse_to_csv(target_path, node_label_matrix, separator="\t", directed=True, numbering="matlab")



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import argparse

from reveal_user_classification.preprocess.snow.snow_2014_graph_dataset_util import process_tweet_collection,\
    make_directory_tree, weakly_connected_graph, make_implicit_graphs, make_annotation
from reveal_user_annotation.mongo.store_snow_data import extract_all_snow_tweets_from_disk_generator
from reveal_user_annotation.mongo.preprocess_data import get_collection_documents_generator
from reveal_user_annotation.mongo.mongo_util import establish_mongo_connection
from reveal_user_annotation.mongo.store_snow_data import store_snow_tweets_from_disk_to_mongodb


def main():
    # Parse arguments.
    parser = argparse.ArgumentParser()
    parser.add_argument("-stf", "--snow-tweets-folder", dest="snow_tweets_folder",
                        help="This is the folder with the SNOW tweets.",
                        type=str, required=True)
    parser.add_argument("-gdf", "--graph-dataset-folder", dest="graph_dataset_folder",
                        help="This is the root directory that the graph dataset will be extracted.",
                        type=str, required=True)
    parser.add_argument("-tlf", "--twitter-lists-folder", dest="twitter_lists_folder",
                        help="This is the folder where the twitter lists for each user are stored.",
                        type=str, required=True)
    parser.add_argument("-tlkf", "--twitter-lists-keywords-folder", dest="twitter_lists_keywords_folder",
                        help="This is the folder where the extracted keywords from the lists for each user are stored.",
                        type=str, required=True)

    args = parser.parse_args()

    snow_tweets_folder = args.snow_tweets_folder
    graph_dataset_folder = args.graph_dataset_folder
    twitter_lists_folder = args.twitter_lists_folder
    twitter_lists_keywords_folder = args.twitter_lists_keywords_folder

    make_dataset(snow_tweets_folder, graph_dataset_folder, twitter_lists_folder, twitter_lists_keywords_folder)


def make_dataset(snow_tweets_folder, graph_dataset_folder, twitter_lists_folder, twitter_lists_keywords_folder):
    # Get a generator of the SNOW 2014 Data Challenge tweets.
    # store_snow_tweets_from_disk_to_mongodb(snow_tweets_folder)

    # Write all tweets on MongoDB.
    # mongo_client = establish_mongo_connection("mongodb://admin:123456@160.40.51.29:27017")

    # Read all tweets in ascending timestamp order.
    # tweet_generator = get_collection_documents_generator(client=mongo_client,
    #                                                      database_name="snow_tweet_storage",
    #                                                      collection_name="tweets",
    #                                                      spec=None,
    #                                                      latest_n=None,
    #                                                      sort_key="created_at")

    # Make sub-folders for the graph dataset.
    full_graph_folder, weakly_connected_graph_folder, weakly_connected_label_folder, implicit_graph_folder,\
    simple_undirected_graph_folder, combinatorial_implicit_graph_folder,\
           directed_implicit_graph_folder = make_directory_tree(graph_dataset_folder)

    # Extract the full graphs from the tweet collection. This is a quadratic complexity process.
    # tweet_generator = extract_all_snow_tweets_from_disk_generator(snow_tweets_folder)
    # process_tweet_collection(tweet_generator, full_graph_folder)

    # Extract weakly connected mention graph and corresponding retweet graph and user_lemma_matrix.
    weakly_connected_graph(full_graph_folder, weakly_connected_graph_folder)

    # Make combinatorial and directed implicit graphs for the mention and retweet graphs.
    make_implicit_graphs(weakly_connected_graph_folder,
                         simple_undirected_graph_folder,
                         combinatorial_implicit_graph_folder,
                         directed_implicit_graph_folder)

    # Make annotation for weakly connected mention graph.
    # make_annotation(twitter_lists_folder,
    #                 twitter_lists_keywords_folder,
    #                 weakly_connected_graph_folder,
    #                 weakly_connected_label_folder,
    #                 full_graph_folder)


# make_dataset(snow_tweets_folder="/home/georgerizos/Documents/LocalStorage/raw_data/SNOW/snow_tweets_folder",
#              graph_dataset_folder="/home/georgerizos/Documents/LocalStorage/raw_data/SNOW",
#              twitter_lists_folder="/home/georgerizos/Documents/LocalStorage/raw_data/SNOW/twitter_lists",
#              twitter_lists_keywords_folder="/home/georgerizos/Documents/LocalStorage/raw_data/SNOW/twitter_lists_keywords")



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from reveal_user_annotation.common.config_package import get_threads_number

from reveal_user_annotation.rabbitmq.rabbitmq_util import establish_rabbitmq_connection, simple_notification,\
    rabbitmq_server_service
from reveal_user_classification.reveal.utility import make_time_window_filter, safe_establish_mongo_connection,\
    get_graphs_and_lemma_matrix, integrate_graphs, fetch_twitter_lists, annotate_users, user_classification,\
    get_user_topic_generator, write_results_to_mongo, write_topics_to_pserver, publish_results_via_rabbitmq


def user_network_profile_classifier(mongo_uri,
                                    assessment_id,
                                    twitter_app_key,
                                    twitter_app_secret,
                                    rabbitmq_uri,
                                    rabbitmq_queue,
                                    rabbitmq_exchange,
                                    rabbitmq_routing_key,
                                    pserver_host_name,
                                    pserver_client_name,
                                    pserver_client_pass,
                                    latest_n,
                                    lower_timestamp,
                                    upper_timestamp,
                                    restart_probability,
                                    number_of_threads,
                                    number_of_users_to_annotate,
                                    max_number_of_labels,
                                    user_network_profile_classifier_db,
                                    local_resources_folder):
    """
    Performs Online Social Network user classification.

    Specifically:
           - Establishes a connection with a Mongo database and gets the newest tweets.
           - Forms graphs and text-based vector representation for the users involved.
           - Fetches Twitter lists for influential users.
           - Extracts keywords from Twitter lists and thus annotates these users as experts in these topics.
           - Extracts graph-based features using the ARCTE algorithm.
           - Performs user classification for the rest of the users.
           - Stores the results at PServer.

    Input: - mongo_uri: A mongo client URI.
           - assessment_id:
           - twitter_app_key:
           - twitter_app_secret:
           - rabbitmq_uri:
           - rabbitmq_queue:
           - rabbitmq_exchange:
           - rabbitmq_routing_key:
           - pserver_host_name:
           - pserver_client_name:
           - pserver_client_pass:
           - latest_n: Get only the N most recent documents.
           - lower_timestamp: Get only documents created after this UNIX timestamp.
           - upper_timestamp: Get only documents created before this UNIX timestamp.
           - restart_probability:
           - number_of_threads:
           - number_of_users_to_annotate:
           - max_number_of_labels:
           - user_network_profile_classifier_db:
           - local_resources_folder: The preprocessed Twitter lists for a number of users are stored here.
    """
    ####################################################################################################################
    # Manage argument input.
    ####################################################################################################################
    spec = make_time_window_filter(lower_timestamp, upper_timestamp)

    if number_of_threads is None:
        number_of_threads = get_threads_number()

    ####################################################################################################################
    # Establish MongoDB connection.
    ####################################################################################################################
    client,\
    tweet_input_database_name,\
    tweet_input_collection_name = safe_establish_mongo_connection(mongo_uri, assessment_id)
    print("MongoDB connection established.")

    ####################################################################################################################
    # Preprocess tweets.
    ####################################################################################################################
    mention_graph,\
    retweet_graph,\
    user_id_set,\
    node_to_id = get_graphs_and_lemma_matrix(client,
                                             tweet_input_database_name,
                                             tweet_input_collection_name,
                                             spec,
                                             latest_n)
    print("Users and user interactions extracted.")

    adjacency_matrix,\
    node_to_id,\
    features,\
    node_importances = integrate_graphs(mention_graph,
                                        retweet_graph,
                                        node_to_id,
                                        restart_probability,
                                        number_of_threads)
    number_of_users = adjacency_matrix.shape[0]
    print("Number of users in biggest network component: ", number_of_users)
    if number_of_users < 2:
        rabbitmq_server_service("restart")
        rabbitmq_connection = establish_rabbitmq_connection(rabbitmq_uri)

        simple_notification(rabbitmq_connection, rabbitmq_queue, rabbitmq_exchange, rabbitmq_routing_key, "FAILURE")
        rabbitmq_connection.close()
        print("Failure message published via RabbitMQ.")

    ####################################################################################################################
    # Annotate users.
    ####################################################################################################################
    twitter_lists_gen,\
    user_ids_to_annotate,\
    user_twitter_ids_mongo,\
    user_twitter_ids_local = fetch_twitter_lists(client,
                                                 twitter_app_key,
                                                 twitter_app_secret,
                                                 user_network_profile_classifier_db,
                                                 local_resources_folder,
                                                 node_importances,
                                                 number_of_users_to_annotate,
                                                 node_to_id)
    print("Annotating users with Twitter ids: ", user_ids_to_annotate)

    user_label_matrix,\
    annotated_user_ids,\
    label_to_lemma,\
    lemma_to_keyword = annotate_users(client,
                                      twitter_lists_gen,
                                      user_ids_to_annotate,
                                      user_twitter_ids_mongo,
                                      user_twitter_ids_local,
                                      local_resources_folder,
                                      user_network_profile_classifier_db,
                                      node_to_id,
                                      max_number_of_labels)
    number_of_labels = user_label_matrix.shape[1]
    print("Number of labels for classification: ", number_of_labels)
    if number_of_labels < 2:
        rabbitmq_server_service("restart")
        rabbitmq_connection = establish_rabbitmq_connection(rabbitmq_uri)

        simple_notification(rabbitmq_connection, rabbitmq_queue, rabbitmq_exchange, rabbitmq_routing_key, "FAILURE")
        rabbitmq_connection.close()
        print("Failure message published via RabbitMQ.")

    ####################################################################################################################
    # Perform user classification.
    ####################################################################################################################
    prediction = user_classification(features, user_label_matrix, annotated_user_ids, node_to_id, number_of_threads)
    print("User classification complete.")

    ####################################################################################################################
    # Write to Mongo, PServer and/or RabbitMQ.
    ####################################################################################################################
    # Write data to mongo.
    write_results_to_mongo(client,
                           user_network_profile_classifier_db,
                           get_user_topic_generator(prediction,
                                                    node_to_id,
                                                    label_to_lemma,
                                                    lemma_to_keyword))
    print("Results written in MongoDB.")

    # Write data to pserver.
    if pserver_host_name is not None:
        topic_list = list(lemma_to_keyword.values())
        try:
            write_topics_to_pserver(pserver_host_name,
                                    pserver_client_name,
                                    pserver_client_pass,
                                    get_user_topic_generator(prediction,
                                                             node_to_id,
                                                             label_to_lemma,
                                                             lemma_to_keyword),
                                    topic_list)
            print("Results written in PServer.")
        except Exception:
            print("Unable to write results to PServer.")

    # Publish results and success message on RabbitMQ.
    rabbitmq_server_service("restart")
    rabbitmq_connection = establish_rabbitmq_connection(rabbitmq_uri)

    publish_results_via_rabbitmq(rabbitmq_connection,
                                 rabbitmq_queue,
                                 rabbitmq_exchange,
                                 rabbitmq_routing_key,
                                 get_user_topic_generator(prediction,
                                                          node_to_id,
                                                          label_to_lemma,
                                                          lemma_to_keyword))
    print("Results published via RabbitMQ.")

    simple_notification(rabbitmq_connection, rabbitmq_queue, rabbitmq_exchange, rabbitmq_routing_key, "SUCCESS")
    rabbitmq_connection.close()
    print("Success message published via RabbitMQ.")



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from setuptools import setup


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name='reveal-user-classification',
    version='0.2.2',
    author='Georgios Rizos',
    author_email='georgerizos@iti.gr',
    packages=['reveal_user_classification',
              'reveal_user_classification.entry_points',
              'reveal_user_classification.preprocess',
              'reveal_user_classification.preprocess.insight',
              'reveal_user_classification.preprocess.snow',
              'reveal_user_classification.reveal'],
    url='https://github.com/MKLab-ITI/reveal-user-classification',
    license='Apache',
    description='Performs user classification into labels using a set of seed Twitter users with known labels and'
                'the structure of the interaction network between them.',
    long_description=readme(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    keywords="online-social-network user-classification Reveal-FP7",
    entry_points={
        'console_scripts': ['user_network_profile_classifier=reveal_user_classification.entry_points.user_network_profile_classifier:main',
                            'prototype_user_network_profile_classifier=reveal_user_classification.entry_points.prototype_user_network_profile_classifier:main',
                            'make_snow2014_graph_dataset=reveal_user_classification.preprocess.snow.make_snow_2014_graph_dataset:main'],
    },
    include_package_data=False,
    install_requires=open("requirements.txt").read().split("\n")
)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import gc
import os
import json
from multiprocessing import Pool
from functools import partial
import numpy as np
import scipy.sparse as spsp
import networkx as nx

from reveal_user_annotation.common.config_package import get_threads_number
from reveal_user_annotation.common.datarw import store_pickle, load_pickle
from reveal_user_annotation.mongo.preprocess_data import extract_graphs_and_lemmas_from_tweets,\
    extract_connected_components
from reveal_user_annotation.text.map_data import chunks
from reveal_user_annotation.text.clean_text import clean_single_word
from reveal_user_annotation.twitter.clean_twitter_list import user_twitter_list_bag_of_words
from reveal_user_annotation.twitter.manage_resources import get_reveal_set, get_topic_keyword_dictionary
from reveal_user_annotation.twitter.user_annotate import form_user_term_matrix, form_lemma_tokeyword_map, filter_user_term_matrix
from reveal_graph_embedding.datautil.snow_datautil import read_adjacency_matrix, scipy_sparse_to_csv,\
    write_screen_name_to_topics
from reveal_user_classification.datautil.make_directory_tree import make_sure_path_exists
from reveal_user_classification.embedding.implicit import get_adjacency_matrix_via_combinatorial_laplacian,\
    get_adjacency_matrix_via_directed_laplacian, get_multiview_transition_matrix, get_implicit_adjacency_matrices


def submatrix_pull_via_networkx(matrix, node_array, directed=True):

    if directed:
        graph = nx.from_scipy_sparse_matrix(matrix, create_using=nx.DiGraph())
    else:
        graph = nx.from_scipy_sparse_matrix(matrix, create_using=nx.Graph())

    sub_graph = graph.subgraph(list(node_array))

    sub_matrix = nx.to_scipy_sparse_matrix(sub_graph, dtype=np.float64, format="csr")

    return sub_matrix


def coo_submatrix_pull(matrix, row_array, col_array):
    # Make sure we are dealing with a coordinate sparse format matrix.
    if type(matrix) != spsp.coo_matrix:
        raise TypeError("Matrix must be sparse COOrdinate format")

    # Initialize mask index arrays.
    gr = -1 * np.ones(matrix.shape[0])
    gc = -1 * np.ones(matrix.shape[1])

    submatrix_row_size = row_array.size
    submatrix_col_size = col_array.size
    ar = np.arange(0, submatrix_row_size)
    ac = np.arange(0, submatrix_col_size)
    gr[row_array[ar]] = ar
    gc[col_array[ac]] = ac

    mrow = matrix.row
    mcol = matrix.col
    newelem = (gr[mrow] > -1) & (gc[mcol] > -1)
    newrows = mrow[newelem]
    newcols = mcol[newelem]

    submatrix = spsp.coo_matrix((matrix.data[newelem], np.array([gr[newrows], gc[newcols]])),
                                shape=(submatrix_row_size, submatrix_col_size))

    return submatrix


def make_directory_tree(graph_dataset_folder):
    full_graph_folder = graph_dataset_folder + "/full_graph"
    weakly_connected_graph_folder = graph_dataset_folder + "/weakly_connected_graph"
    weakly_connected_label_folder = graph_dataset_folder + "/weakly_connected_graph/labels"
    implicit_graph_folder = weakly_connected_graph_folder + "/implicit_graph"
    simple_undirected_graph_folder = implicit_graph_folder + "/simple_undirected_implicit_graph"
    combinatorial_implicit_graph_folder = implicit_graph_folder + "/combinatorial_implicit_graph"
    directed_implicit_graph_folder = implicit_graph_folder + "/directed_implicit_graph"

    make_sure_path_exists(full_graph_folder)
    make_sure_path_exists(weakly_connected_graph_folder)
    make_sure_path_exists(weakly_connected_label_folder)
    make_sure_path_exists(implicit_graph_folder)
    make_sure_path_exists(simple_undirected_graph_folder)
    make_sure_path_exists(combinatorial_implicit_graph_folder)
    make_sure_path_exists(directed_implicit_graph_folder)

    return full_graph_folder, weakly_connected_graph_folder, weakly_connected_label_folder, implicit_graph_folder,\
           simple_undirected_graph_folder, combinatorial_implicit_graph_folder, directed_implicit_graph_folder


def process_tweet_collection(tweet_generator, full_graph_folder):
    mention_graph,\
    retweet_graph,\
    user_lemma_matrix,\
    tweet_id_set,\
    user_id_set,\
    node_to_id,\
    lemma_to_attribute,\
    id_to_name = extract_graphs_and_lemmas_from_tweets(tweet_generator)

    # Store full graph data in corresponding folder.
    store_pickle(full_graph_folder + "/mention_graph" + ".pkl", mention_graph)
    scipy_sparse_to_csv(full_graph_folder + "/mention_graph" + ".tsv", mention_graph, "\t", directed=True)
    store_pickle(full_graph_folder + "/retweet_graph" + ".pkl", retweet_graph)
    scipy_sparse_to_csv(full_graph_folder + "/retweet_graph" + ".tsv", retweet_graph, "\t", directed=True)
    store_pickle(full_graph_folder + "/user_lemma_matrix" + ".pkl", user_lemma_matrix)
    scipy_sparse_to_csv(full_graph_folder + "/user_lemma_matrix" + ".tsv", user_lemma_matrix, "\t", directed=True)
    store_pickle(full_graph_folder + "/tweet_id_set" + ".pkl", tweet_id_set)
    store_pickle(full_graph_folder + "/user_id_set" + ".pkl", user_id_set)
    store_pickle(full_graph_folder + "/node_to_id" + ".pkl", node_to_id)
    store_pickle(full_graph_folder + "/lemma_to_attribute" + ".pkl", lemma_to_attribute)
    store_pickle(full_graph_folder + "/id_to_name" + ".pkl", id_to_name)


def weakly_connected_graph(full_graph_folder, weakly_connected_graph_folder):
    # Read relevant data.
    mention_graph = load_pickle(full_graph_folder + "/mention_graph" + ".pkl")
    mention_graph = spsp.coo_matrix(spsp.csr_matrix(mention_graph))
    retweet_graph = load_pickle(full_graph_folder + "/retweet_graph" + ".pkl")
    retweet_graph = spsp.coo_matrix(spsp.csr_matrix(retweet_graph))
    user_lemma_matrix = load_pickle(full_graph_folder + "/user_lemma_matrix" + ".pkl")
    user_lemma_matrix = spsp.coo_matrix(spsp.csr_matrix(user_lemma_matrix))
    user_id_set = load_pickle(full_graph_folder + "/user_id_set" + ".pkl")
    node_to_id = load_pickle(full_graph_folder + "/node_to_id" + ".pkl")

    # Extract weakly connected graph for the mention graph.
    weakly_connected_men_ret_graph, weakly_connected_node_to_id, old_node_list = extract_connected_components(spsp.coo_matrix(spsp.csr_matrix(mention_graph + retweet_graph)),
                                                                                                              "weak",
                                                                                                              node_to_id)

    # Calculate the user twitter id set for the weakly connected component.
    weakly_connected_user_id_set = set(list(weakly_connected_node_to_id.values()))

    node_array = np.array(old_node_list, dtype=np.int64)

    # Extract corresponding retweet graph and user lemma matrix.
    weakly_connected_mention_graph = submatrix_pull_via_networkx(spsp.coo_matrix(mention_graph),
                                                                 node_array,
                                                                 directed=True)

    weakly_connected_retweet_graph = submatrix_pull_via_networkx(spsp.coo_matrix(retweet_graph),
                                                                 node_array,
                                                                 directed=True)

    user_lemma_matrix = spsp.csr_matrix(user_lemma_matrix)
    weakly_connected_user_lemma_matrix = user_lemma_matrix[node_array, :]

    # Change sparse matrices to coordinate format in order to save as an edge list.
    weakly_connected_mention_graph = spsp.coo_matrix(weakly_connected_mention_graph)
    weakly_connected_retweet_graph = spsp.coo_matrix(weakly_connected_retweet_graph)
    weakly_connected_user_lemma_matrix = spsp.coo_matrix(weakly_connected_user_lemma_matrix)

    # Store weakly connected data.
    scipy_sparse_to_csv(weakly_connected_graph_folder + "/mention_graph.tsv",
                        weakly_connected_mention_graph,
                        separator="\t",
                        directed=True)

    scipy_sparse_to_csv(weakly_connected_graph_folder + "/retweet_graph.tsv",
                        weakly_connected_retweet_graph,
                        separator="\t",
                        directed=True)

    scipy_sparse_to_csv(weakly_connected_graph_folder + "/user_lemma_matrix.tsv",
                        weakly_connected_user_lemma_matrix,
                        separator="\t",
                        directed=True)

    store_pickle(weakly_connected_graph_folder + "/user_id_set" + ".pkl", weakly_connected_user_id_set)
    store_pickle(weakly_connected_graph_folder + "/node_to_id" + ".pkl", weakly_connected_node_to_id)


def make_implicit_graphs(weakly_connected_graph_folder,
                         simple_undirected_graph_folder,
                         combinatorial_implicit_graph_folder,
                         directed_implicit_graph_folder):
    # Read relevant data.
    mention_graph = read_adjacency_matrix(weakly_connected_graph_folder + "/mention_graph.tsv", separator="\t")
    retweet_graph = read_adjacency_matrix(weakly_connected_graph_folder + "/retweet_graph.tsv", separator="\t")
    # user_lemma_matrix = read_adjacency_matrix(weakly_connected_graph_folder + "/user_lemma_matrix.tsv", separator="\t")

    # Make text-based graph.
    # lemma_graph = make_text_graph(user_lemma_matrix)

    ####################################################################################################################
    # Make simple undirected graphs.
    ####################################################################################################################
    simple_undirected_mention_graph = (mention_graph + mention_graph.transpose())/2
    simple_undirected_mention_graph = spsp.coo_matrix(spsp.csr_matrix(simple_undirected_mention_graph))
    scipy_sparse_to_csv(simple_undirected_graph_folder + "/mention_graph" + ".tsv",
                        simple_undirected_mention_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Simple Undirected Mention Graph.")

    simple_undirected_retweet_graph = (retweet_graph + retweet_graph.transpose())/2
    simple_undirected_retweet_graph = spsp.coo_matrix(spsp.csr_matrix(simple_undirected_retweet_graph))
    scipy_sparse_to_csv(simple_undirected_graph_folder + "/retweet_graph" + ".tsv",
                        simple_undirected_retweet_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Simple Undirected Retweet Graph.")

    # simple_undirected_lemma_graph = (lemma_graph + lemma_graph.transpose())/2
    # simple_undirected_lemma_graph = spsp.coo_matrix(spsp.csr_matrix(simple_undirected_lemma_graph))
    # scipy_sparse_to_csv(simple_undirected_graph_folder + "/lemma_graph" + ".tsv",
    #                     simple_undirected_lemma_graph,
    #                     separator="\t",
    #                     directed=False)
    # gc.collect()
    # print("Simple Undirected Lemma Graph.")

    simple_undirected_mr_graph = (simple_undirected_mention_graph + simple_undirected_retweet_graph)/2
    simple_undirected_mr_graph = spsp.coo_matrix(spsp.csr_matrix(simple_undirected_mr_graph))
    scipy_sparse_to_csv(simple_undirected_graph_folder + "/men_ret_graph" + ".tsv",
                        simple_undirected_mr_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Simple Undirected Mention+Retweet Graph.")

    ####################################################################################################################
    # Make combinatorial implicit graphs.
    ####################################################################################################################
    implicit_combinatorial_mention_graph, phi = get_adjacency_matrix_via_combinatorial_laplacian(mention_graph, 0.1)
    implicit_combinatorial_mention_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_combinatorial_mention_graph))
    scipy_sparse_to_csv(combinatorial_implicit_graph_folder + "/mention_graph" + ".tsv",
                        implicit_combinatorial_mention_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Implicit Combinatorial Mention Graph.")
    print(implicit_combinatorial_mention_graph.sum(axis=1))

    implicit_combinatorial_retweet_graph, phi = get_adjacency_matrix_via_combinatorial_laplacian(retweet_graph, 0.1)
    implicit_combinatorial_retweet_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_combinatorial_retweet_graph))
    scipy_sparse_to_csv(combinatorial_implicit_graph_folder + "/retweet_graph" + ".tsv",
                        implicit_combinatorial_retweet_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Implicit Combinatorial Retweet Graph.")
    print(implicit_combinatorial_retweet_graph.sum(axis=1))

    # implicit_combinatorial_lemma_graph, phi = get_adjacency_matrix_via_combinatorial_laplacian(lemma_graph, 0.5)
    # implicit_combinatorial_lemma_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_combinatorial_lemma_graph))
    # scipy_sparse_to_csv(combinatorial_implicit_graph_folder + "/lemma_graph" + ".tsv",
    #                     implicit_combinatorial_lemma_graph,
    #                     separator="\t",
    #                     directed=False)
    # gc.collect()
    # print("Implicit Combinatorial Lemma Graph.")

    ####################################################################################################################
    # Make and store directed implicit graphs.
    ####################################################################################################################
    implicit_directed_mention_graph, phi = get_adjacency_matrix_via_directed_laplacian(mention_graph, 0.1)
    implicit_directed_mention_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_directed_mention_graph))
    scipy_sparse_to_csv(directed_implicit_graph_folder + "/mention_graph" + ".tsv",
                        implicit_directed_mention_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Implicit Directed Mention Graph.")
    print(implicit_directed_mention_graph.sum(axis=1))

    implicit_directed_retweet_graph, phi = get_adjacency_matrix_via_directed_laplacian(retweet_graph, 0.1)
    implicit_directed_retweet_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_directed_retweet_graph))
    scipy_sparse_to_csv(directed_implicit_graph_folder + "/retweet_graph" + ".tsv",
                        implicit_directed_retweet_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Implicit Directed Retweet Graph.")
    print(implicit_directed_retweet_graph.sum(axis=1))

    # implicit_directed_lemma_graph, phi = get_adjacency_matrix_via_directed_laplacian(lemma_graph, 0.1)
    # implicit_directed_lemma_graph = spsp.coo_matrix(spsp.csr_matrix(implicit_directed_lemma_graph))
    # scipy_sparse_to_csv(directed_implicit_graph_folder + "/lemma_graph" + ".tsv",
    #                     implicit_directed_lemma_graph,
    #                     separator="\t",
    #                     directed=False)
    # gc.collect()
    # print("Implicit Directed Lemma Graph.")

    ####################################################################################################################
    # Make multiview transition matrices.
    ####################################################################################################################
    men_ret_transition_matrix = get_multiview_transition_matrix([mention_graph,
                                                                 retweet_graph],
                                                                weights=None,
                                                                method="zhou")

    implicit_combinatorial_men_ret_graph, com_phi,\
    implicit_directed_men_ret_graph, dir_phi = get_implicit_adjacency_matrices(men_ret_transition_matrix,
                                                                               rho=0.1)
    scipy_sparse_to_csv(combinatorial_implicit_graph_folder + "/men_ret_graph" + ".tsv",
                        implicit_combinatorial_men_ret_graph,
                        separator="\t",
                        directed=False)
    scipy_sparse_to_csv(directed_implicit_graph_folder + "/men_ret_graph" + ".tsv",
                        implicit_directed_men_ret_graph,
                        separator="\t",
                        directed=False)
    gc.collect()
    print("Implicit Mention-Retweet Graphs.")
    # men_lem_transition_matrix = get_multiview_transition_matrix([mention_graph,
    #                                                              lemma_graph],
    #                                                             weights=None,
    #                                                             method="zhou")
    # implicit_combinatorial_men_lem_graph, com_phi,\
    # implicit_directed_men_lem_graph, dir_phi = get_implicit_adjacency_matrices(men_lem_transition_matrix,
    #                                                                            rho=0.2)
    # gc.collect()
    # print("Implicit Mention-Lemma Graphs.")
    #
    # men_ret_lem_transition_matrix = get_multiview_transition_matrix([mention_graph,
    #                                                                  retweet_graph,
    #                                                                  lemma_graph],
    #                                                                 weights=None,
    #                                                                 method="zhou")
    # implicit_combinatorial_men_ret_lem_graph, com_phi,\
    # implicit_directed_men_ret_lem_graph, dir_phi = get_implicit_adjacency_matrices(men_ret_lem_transition_matrix,
    #                                                                                rho=0.2)
    # gc.collect()
    # print("Implicit Mention-Retweet-Lemma Graphs.")


def make_annotation(twitter_lists_folder, twitter_lists_keywords_folder, weakly_connected_graph_folder, weakly_connected_label_folder, full_graph_folder):
    # TODO: Move keywords from Mongo to the folder.
    # Read set of users.
    weakly_connected_user_id_set = load_pickle(weakly_connected_graph_folder + "/user_id_set" + ".pkl")
    weakly_connected_node_to_id = load_pickle(weakly_connected_graph_folder + "/node_to_id" + ".pkl")
    id_to_name = load_pickle(full_graph_folder + "/id_to_name" + ".pkl")

    # Read set of twitter lists.
    twitter_list_file_list = os.listdir(twitter_lists_folder)
    twitter_list_file_list = [int(file_name[:-4]) for file_name in twitter_list_file_list]

    # Read which users are annotated.
    user_keywords_file_list = os.listdir(twitter_lists_keywords_folder)
    user_keywords_file_list = [int(file_name[:-5]) for file_name in user_keywords_file_list]

    # Find which twitter lists need to be preprocessed.
    user_twitter_id_list = [file_name for file_name in twitter_list_file_list if file_name in weakly_connected_user_id_set]
    user_twitter_id_list = [file_name for file_name in user_twitter_id_list if file_name not in user_keywords_file_list]

    twitter_list_file_list = [str(file_name) + ".pkl" for file_name in user_twitter_id_list]

    pool = Pool(processes=get_threads_number()*2,)
    user_chunks = chunks(twitter_list_file_list, get_threads_number()*2)
    pool.map(partial(worker_function,
                     lemmatizing="wordnet",
                     source_folder=twitter_lists_folder,
                     target_folder=twitter_lists_keywords_folder),
             user_chunks)

    # # Make user-label matrix.
    user_keywords_file_list = [str(file_name) for file_name in user_keywords_file_list]
    user_twitter_list_keywords_gen = read_local_user_annotations(twitter_lists_keywords_folder,
                                                                 user_keywords_file_list)
    weakly_connected_id_to_node = dict(zip(weakly_connected_node_to_id.values(),
                                           weakly_connected_node_to_id.keys()))

    # # twitter_id_to_weakly_connected_node = {int(twitter_id): weakly_connected_id_to_node[int(twitter_id)] for twitter_id in user_keywords_file_list if int(twitter_id) in weakly_connected_id_to_node.keys()}
    # node_twitter_list_keywords_gen = ((weakly_connected_id_to_node[int(user_twitter_id)], twitter_list_keywords) for user_twitter_id, twitter_list_keywords in user_twitter_list_keywords_gen if int(user_twitter_id) in weakly_connected_id_to_node.keys())
    # for node, j in user_twitter_list_keywords_gen:
    #     print(node, j)

    implicated_user_twitter_list_keywords_gen = ((int(user_twitter_id), twitter_list_keywords) for user_twitter_id, twitter_list_keywords in user_twitter_list_keywords_gen if int(user_twitter_id) in weakly_connected_id_to_node.keys())
    # for node, j in user_twitter_list_keywords_gen:
    #     print(node, j)

    ####################################################################################################################
    # Semi-automatic user annotation.
    ####################################################################################################################
    reveal_set = get_reveal_set()
    topic_keyword_dict = get_topic_keyword_dictionary()

    available_topics = set(list(topic_keyword_dict.keys()))

    keyword_list = list()
    for topic in reveal_set:
        if topic in available_topics:
            keyword_list.extend(topic_keyword_dict[topic])

    lemma_set = list()
    for keyword in keyword_list:
        lemma = clean_single_word(keyword, lemmatizing="wordnet")
        lemma_set.append(lemma)
    lemma_set = set(lemma_set)

    keyword_topic_dict = dict()
    for topic, keyword_set in topic_keyword_dict.items():
        for keyword in keyword_set:
            keyword_topic_dict[keyword] = topic

    user_label_matrix, annotated_nodes, label_to_lemma, node_to_lemma_tokeywordbag = form_user_term_matrix(implicated_user_twitter_list_keywords_gen,
                                                                                                           weakly_connected_id_to_node,
                                                                                                           lemma_set=lemma_set,
                                                                                                           keyword_to_topic_manual=keyword_topic_dict)

    scipy_sparse_to_csv(weakly_connected_label_folder + "/unfiltered_user_label_matrix" + ".tsv",
                        user_label_matrix,
                        "\t",
                        directed=True)
    store_pickle(weakly_connected_label_folder + "/unfiltered_annotated_nodes" + ".pkl",
                 annotated_nodes)
    store_pickle(weakly_connected_label_folder + "/unfiltered_label_to_lemma" + ".pkl",
                 label_to_lemma)
    store_pickle(weakly_connected_label_folder + "/unfiltered_node_to_lemma_tokeywordbag" + ".pkl",
                 node_to_lemma_tokeywordbag)


    user_label_matrix, annotated_user_ids, label_to_lemma = filter_user_term_matrix(user_label_matrix,
                                                                                    annotated_nodes,
                                                                                    label_to_lemma,
                                                                                    max_number_of_labels=None)

    lemma_to_keyword = form_lemma_tokeyword_map(annotated_nodes, node_to_lemma_tokeywordbag)

    # user_label_matrix, annotated_user_ids, label_to_lemma, lemma_to_keyword = semi_automatic_user_annotation(implicated_user_twitter_list_keywords_gen, weakly_connected_id_to_node)

    # Store user-label binary matrix.
    scipy_sparse_to_csv(weakly_connected_label_folder + "/user_label_matrix" + ".tsv",
                        user_label_matrix,
                        "\t",
                        directed=True)

    # Store user-label keyword matrix.
    write_screen_name_to_topics(weakly_connected_label_folder + "/user_name_to_topics" + ".tsv",
                                user_label_matrix,
                                weakly_connected_node_to_id,
                                id_to_name,
                                label_to_lemma,
                                lemma_to_keyword,
                                separator="\t")
    return twitter_lists_folder


def worker_function(file_name_list,
                    lemmatizing,
                    source_folder,
                    target_folder):
    source_path_list = (source_folder + "/" + file_name for file_name in file_name_list)
    target_path_list = (target_folder + "/" + file_name[:-4] + ".json" for file_name in file_name_list)

    # Get the lists of a user
    for source_path in source_path_list:
        twitter_lists_corpus = load_pickle(source_path)
        if "lists" in twitter_lists_corpus.keys():
            twitter_lists_corpus = twitter_lists_corpus["lists"]
        else:
            continue

        bag_of_lemmas, lemma_to_keywordbag = user_twitter_list_bag_of_words(twitter_lists_corpus, lemmatizing)

        user_annotation = dict()
        user_annotation["bag_of_lemmas"] = bag_of_lemmas
        user_annotation["lemma_to_keywordbag"] = lemma_to_keywordbag

        target_path = next(target_path_list)
        with open(target_path, "w", encoding="utf-8") as fp:
            json.dump(user_annotation, fp)


def read_local_user_annotations(json_folder,
                                user_twitter_ids):
    if json_folder is not None:
        for user_twitter_id in user_twitter_ids:
            path = json_folder + "/" + str(user_twitter_id) + ".json"
            with open(path, "r", encoding="utf-8") as f:
                twitter_lists = json.load(f)

                yield user_twitter_id, twitter_lists
    else:
        raise StopIteration
